package str.classtree.atermify;

import java.util.*;
import org.apache.bcel.classfile.*;
import org.apache.bcel.generic.*;
import aterm.*;
import static str.classtree.atermify.ATermifyUtil.*;
import str.aterm.ATermifier;
import static str.aterm.ATermUtil.*;

public class ClassATermifier implements ATermifier<JavaClass>
{
  public ATerm convert(JavaClass clazz)
  {
    ATermifier<Field> afield = new FieldATermifier();
    ATermifier<Method> amethod = new MethodATermifier(clazz.getConstantPool());
    ClassAttributeATermifier aattrs = new ClassAttributeATermifier(clazz);

    ATermList fieldsATerm = _factory.makeList();
    Field[] fields = clazz.getFields();
    for(int i = fields.length - 1; i >= 0; i--)
    {
      fieldsATerm = fieldsATerm.insert(afield.convert(fields[i]));
    }
            
    ATermList methodsATerm = _factory.makeList();
    Method[] methods = clazz.getMethods();
    for(int i = methods.length - 1; i >= 0; i--)
    {
      methodsATerm = methodsATerm.insert(amethod.convert(methods[i]));
    }

    return appl("ClassFile",
      appl("MinorVersion", aint(clazz.getMinor())),
      appl("MajorVersion", aint(clazz.getMajor())),
      appl("AccessFlags", buildFlags(clazz)),
      appl("ThisClass", astring(clazz.getClassName())),
      appl("SuperClass", buildSuperClass(clazz)),
      appl("Interfaces", buildInterfaces(clazz)),
      appl("Fields", fieldsATerm),
      appl("Methods", methodsATerm),
      appl("Attributes", aattrs.map(clazz.getAttributes()))
    );
  }

  /**
   * If the SuperclassNameIndex is 0, then the current class is the
   * java.lang.Object class. This is the only class without a super
   * class.
   */
  private ATerm buildSuperClass(JavaClass clazz)
  {
    if(clazz.getSuperclassNameIndex() == 0)
    {
      return appl("None");
    }
    else
    {
      return appl("Some", astring(clazz.getSuperclassName()));
    }
  }

  private ATerm buildInterfaces(JavaClass clazz)
  {
    ATermList result = _factory.makeList();

    String[] interfaces = clazz.getInterfaceNames();
    for(int i = interfaces.length - 1; i >= 0; i--)
    {
      result = result.insert(astring(interfaces[i]));
    }

    return result;
  }
}
