package str.classtree.atermify;

import java.util.*;
import org.apache.bcel.classfile.*;
import org.apache.bcel.generic.*;
import aterm.*;
import str.classtree.Config;
import static str.classtree.atermify.ATermifyUtil.*;
import static str.aterm.ATermUtil.*;

/**
 * Standard method attributes:
 *
 *   Code
 *   Exceptions
 *   Synthetic
 *   Signature 
 *   Deprecated
 */
public class MethodAttributeATermifier extends AbstractAttributeATermifier
{
  private Method _m;
  private MethodGen _mg;
  private ConstantPool _cp;

  public MethodAttributeATermifier(Method m, ConstantPool cp)
  {
    super();
    _m = m;
    _cp = cp;
  }

  private MethodGen getMethodGen()
  {
    if(_mg == null)
      _mg = new MethodGen(_m, "Voodoo", new ConstantPoolGen(_cp));

    return _mg;
  }

  public ATerm convert(Attribute attr)
  {
    ATerm result;

    if(attr instanceof Code)
    {
      if(Config.disasm_code) 
      {
	result = buildCode((Code) attr);
      }
      else
      {
	result = null;
      }
    }
    else if(attr instanceof Synthetic)
    {
      result = buildSyntheticAttribute((Synthetic) attr);
    }
    else if(attr instanceof ExceptionTable)
    {
      result = buildExceptions((ExceptionTable) attr);
    }
    else if(attr instanceof org.apache.bcel.classfile.Deprecated)
    {
      result = buildDeprecatedAttribute((org.apache.bcel.classfile.Deprecated) attr);
    }
    else if(attr instanceof Signature)
    {
      result = appl("MethodSignature", astring(((Signature) attr).getSignature()));
    }
    else
    {
      result = buildUnknownAttribute(attr);
    }

    return result;
  }

  public ATerm buildExceptions(ExceptionTable table)
  {
    ATermList list = _factory.makeList();
    String[] exceptions = table.getExceptionNames();
    for (int i = exceptions.length - 1; i >= 0; i--)
    {
      list = list.insert(appl("Class", astring(exceptions[i])));
    }

    return appl("Exceptions", list);
  }

  public ATerm buildCode(Code code)
  {
    MethodGen mg = getMethodGen();

    CodeAttributeATermifier acode = new CodeAttributeATermifier(mg);

    ConstantPoolGen cpg     = mg.getConstantPool();
    InstructionList     il  = mg.getInstructionList();
    InstructionHandle[] ihs = il.getInstructionHandles();

    ATermList instructions = _factory.makeList();
    InstructionATermifier atermifier = new InstructionATermifier(cpg);
    for (int i = ihs.length - 1; i >= 0; i--)
    {
      instructions = instructions.insert(atermifier.convert(ihs[i]));
    }

    CodeExceptionGen[] excs = mg.getExceptionHandlers();
    ATermList handlers = _factory.makeList();
    for (int i = excs.length - 1; i >= 0; i--)
    {
      CodeExceptionGen exc = excs[i];
      int start = getInstructionPosition(mg.getInstructionList(), exc.getStartPC());
      int end   = getInstructionPosition(mg.getInstructionList(), exc.getEndPC());
      int handler = getInstructionPosition(mg.getInstructionList(), exc.getHandlerPC());

      ATerm catchtype;
      if(exc.getCatchType() == null)
	catchtype = appl("Any");
      else
	catchtype = TypeATermifier.SHARED.convert(exc.getCatchType());

      handlers = handlers.insert(
	appl("Entry",
	  appl("Range", aint(start), aint(end)),
	  appl("Handler", aint(handler)),
	  appl("CatchType", catchtype)
	)
      );
    }
    return appl("Code",
      appl("MaxStack", appl("Some", aint(code.getMaxStack()))),
      appl("MaxLocals", appl("Some", aint(code.getMaxLocals()))),
      appl("Instructions", instructions),
      appl("ExceptionTable", handlers),
      appl("Attributes", acode.map(code.getAttributes()))
    );
  }

}

